/*
 * Created on May 15, 2006
 * 
 */
package se.umu.wsag.impl;

import java.io.File;
import java.security.Principal;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.security.auth.Subject;
import javax.xml.namespace.QName;
import javax.xml.soap.SOAPBody;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPMessage;
import javax.xml.soap.SOAPPart;

import org.apache.axis.AxisFault;
import org.apache.axis.Message;
import org.apache.axis.MessageContext;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.message.MessageElement;
import org.apache.axis.message.addressing.EndpointReferenceType;
import org.apache.log4j.Logger;
import org.globus.delegation.DelegationUtil;
import org.globus.delegation.service.DelegationResource;
import org.globus.exec.generated.CreateManagedJobInputType;
import org.globus.exec.utils.ManagedJobConstants;
import org.globus.gsi.GlobusCredential;
import org.globus.gsi.jaas.UserNamePrincipal;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.impl.security.authentication.Constants;
import org.globus.wsrf.impl.security.util.AuthUtil;
import org.w3c.dom.NodeList;

/**
 * @author tordsson
 *
 * This class is used for reservation authorization for incoming job requests. 
 * If a job request specifies that the job should run in a reservation, 
 * it is ensured that the reservation i) exists, and ii), is created
 * by the same user that submits the job. Should any of these two 
 * tests fail, the job request is denied. 
 * 
 * This class temporarly replaces the ReservationPDP, as the 
 * ReservationPDP does not work with Globus 4.0.2 due to bug #3897 
 */
public class WSGRAMReservationInterceptor extends BasicHandler {
	
	private static Logger logger = Logger.getLogger(WSGRAMReservationInterceptor.class); 
	
	public static QName CREATE_MANAGED_JOB = new QName(
			ManagedJobConstants.JOB_NS, "createManagedJob"); 
	
	private String reservationProxyLocation = null; 
	
	public void invoke(MessageContext messageContext) throws AxisFault {
		
		QName invokedOperation = messageContext.getOperation().getElementQName(); 
		
		// We are only interested in createManagedJob
		if(invokedOperation.equals(CREATE_MANAGED_JOB)) {
			logger.info("createManagedJob called"); 	
			
			String reservationID = getReservation(messageContext); 
			if(reservationID != null) {
				logger.info("Job specifies reservation : " + reservationID); 
				authorizeReservation(messageContext, reservationID); 
			}
			else {
				logger.debug("No reservation specified"); 
			}
		}
	}
	
	public void authorizeReservation(MessageContext context, 
			String reservationID) throws AxisFault {
		
		// See if there is a proxy stored for this reservation: 							
		String reservationProxyFilePath = this.reservationProxyLocation 
			+  "reservation." 
			+ reservationID
			+ ".proxy";
		
		File f = new File(reservationProxyFilePath); 
		if(!f.exists()) {
			logger.error("No reservation with id " + reservationID + " found"); 
			throw new AxisFault ("Specified reservation " 
					+ reservationID + " does not exist"); 
		}	
		
		// Get caller's credentials: 
	    GlobusCredential submitterCredentials = null;
        String globalUserId = null;
        String localUserId = null;

        // Get caller's grid identity and local user identity
        Subject subject = (Subject) context.getProperty(Constants.PEER_SUBJECT);
        globalUserId = AuthUtil.getIdentity(subject);
        logger.debug("Job request made by: " + globalUserId);
        Iterator principals = subject.getPrincipals().iterator();
        while (principals.hasNext()) {
            Principal principal = (Principal) principals.next();
            if (principal instanceof UserNamePrincipal) {
                localUserId = ((UserNamePrincipal) principal).getName();             
                logger.debug("Local user id: " + localUserId);                
                break;
            }
        }
        
        // Get caller proxy credentials
        CreateManagedJobInputType jobRequest = getJobRequest(context);         
        if(jobRequest == null) {
        	logger.error("Malformed job request"); 
        	throw new AxisFault("Mailformed job request"); 
        }
        EndpointReferenceType jobCredentialEpr = jobRequest.getJob().getJobCredentialEndpoint();
        if (jobCredentialEpr != null) {
            try {
                DelegationResource resource = DelegationUtil.getDelegationResource(
                    jobCredentialEpr
                );
                submitterCredentials = resource.getCredential();
                // Make sure job submitter is not referencing someone else's
                // delegated credentials
                String submitterIdentity = submitterCredentials.getIdentity(); 
                if (!submitterIdentity.equals(globalUserId)) {
                    throw new AxisFault(
                        "The referenced job credentials do not belong to the requester"
                    );
                }
                
            } catch (Exception e) {
                logger.warn("Failed to retrieve delegated job credentials: " + e);
            }
        }
        
        // Finally, ensure that credentials used to create proxy are the same 
        // as those used to submit the job
		// Get DN used to created referred proxy
        try { 
        	GlobusCredential cred = new GlobusCredential(reservationProxyFilePath); 
        	String reservationDN = cred.getIdentity(); 
        	logger.info("User with DN " + reservationDN + " created reservation");
        	if(!globalUserId.equals(reservationDN)) {
        		logger.error("User did not create reservation, job request denied"); 
        		throw new AxisFault("User did not create reservation, job request denied"); 
        	}        	
        	else {
        		logger.info("User authorization ok, job request accepted"); 
        	}
        } catch (Exception e) {
        	logger.error("Could not read proxy file: " + reservationProxyFilePath); 
        	throw new AxisFault("Could not read proxy file " + reservationProxyFilePath); 
        } 
        
	}
	
	
	// Extract MJFS job request from messagecontext
    public CreateManagedJobInputType getJobRequest(MessageContext ctx) {
    	CreateManagedJobInputType jobRequest = null;
            try {
            	
                Message message = ctx.getCurrentMessage();
                SOAPPart soapPart = message.getSOAPPart();
                SOAPEnvelope envelope = soapPart.getEnvelope();
                SOAPBody soapBody = envelope.getBody();
                Iterator bodyElements = soapBody.getChildElements();
                if (bodyElements.hasNext()) {
                    jobRequest = (CreateManagedJobInputType) ObjectDeserializer.toObject(		
                        (SOAPElement) bodyElements.next(), CreateManagedJobInputType.class
                    );
                }
            } catch (Exception e) {
                logger.error("Failed to dig out soap message parameters: " + e);
            }
    	return jobRequest;
        }

	
	// Extract value of 
	public String getReservation(MessageContext context) {
		
		String reservationId = null; 
		try { 
			// Extract job description (reservation id) from soap message
			SOAPMessage message = context.getMessage();		
			SOAPBody body = message.getSOAPBody();
			
			// Get the <extensions>-tag if it exists. 								
			// Rather ugly parsing of soap element
			List allNodes = new LinkedList(); 			
			NodeList l = body.getChildNodes(); 
			for(int i = 0; i < l.getLength(); i++) {				
				MessageElement child = (MessageElement) l.item(i); 
				allNodes.add(child); 
			}
			
			
			boolean found = false; 
			while(allNodes.size() > 0 && !found) {
				MessageElement node = (MessageElement) allNodes.remove(0); 				
				Iterator iter = node.getChildElements(); 
				while(iter.hasNext()) {
					MessageElement child = (MessageElement) iter.next(); 
					if("job".equals(child.getName())) {
						allNodes.add(child); 
					}
					if("extensions".equals(child.getName())) {
						allNodes.add(child); 
					}
					if("ReservationType".equals(child.getName())) {
						reservationId = child.getChildNodes().item(0).toString(); 
						found = true; 
					}
				}
			}
			
			return reservationId; 
		} catch (Exception e) {
			logger.warn("Malformed job request"); 
			return null; 
		}
	}
	
	public void init() {
		this.options = getOptions(); 
		this.reservationProxyLocation = (String) this.getOption(
				"se.umu.wsag.reservationProxyLocation"); 
		logger.info("Reservation proxy location: " + this.reservationProxyLocation); 
	}
	
}
